/*
 * Copyright
 * (C) 2024 Das Land Schleswig-Holstein vertreten durch das
 *  Minis
 * erium für Energiewende, Klimaschutz, Umwelt und Natur Zentrales
 *  IT-Management 
 * 
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europäischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * 
 * 
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 * 
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschränkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 */
package de.ozgcloud.admin.security;

import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import java.net.URI;
import java.util.Collections;
import java.util.Map;

import org.apache.http.client.utils.URIBuilder;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.http.MediaType;
import org.springframework.test.context.DynamicPropertyRegistry;
import org.springframework.test.context.DynamicPropertySource;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;
import org.springframework.web.client.RestClient;

import dasniko.testcontainers.keycloak.KeycloakContainer;
import de.ozgcloud.admin.RootController;
import de.ozgcloud.common.test.DataITCase;
import lombok.SneakyThrows;

@DataITCase
@AutoConfigureMockMvc
class SecurityConfigurationWithKeycloakITCase {
	@Autowired
	private MockMvc mockMvc;

	static KeycloakContainer keycloak;

	@BeforeAll
	static void setupKeycloakContainer() {
		keycloak = new KeycloakContainer().withRealmImportFile("keycloak/realm-export.json");
		keycloak.start();
	}

	@AfterAll
	static void closeKeycloakContainer() {
		keycloak.close();
	}

	@DynamicPropertySource
	static void registerResourceServerIssuerProperty(DynamicPropertyRegistry registry) {
		registry.add("spring.security.oauth2.resourceserver.jwt.issuer-uri", () -> keycloak.getAuthServerUrl() + "/realms/by-kiel-dev");
	}

	@Nested
	class TestSecuredEndpointWithKeycloakToken {
		@SneakyThrows
		@ParameterizedTest
		@ValueSource(strings = {
				"/api/environment",
				"/configserver/name/profile",
				"/api", "/api/configuration", "/api/configuration/param",
		})
		void shouldGetAccessWithToken() {
			String token = getToken();

			var result = mockMvc.perform(get(RootController.PATH).header("Authorization", token));

			result.andExpect(status().isOk());
		}

		@SneakyThrows
		String getToken() {
			MultiValueMap<String, String> formData = setPostBodyForToken();

			Map<String, String> resultBody = performPostRequestToKeycloak(formData);

			return "Bearer " + resultBody.get("access_token").toString();
		}

		MultiValueMap<String, String> setPostBodyForToken() {
			MultiValueMap<String, String> formData = new LinkedMultiValueMap<>();
			formData.put("grant_type", Collections.singletonList("password"));
			formData.put("client_id", Collections.singletonList("admin"));
			formData.put("username", Collections.singletonList("admin-test"));
			formData.put("password", Collections.singletonList("Password"));
			return formData;
		}

		@SuppressWarnings("unchecked")
		@SneakyThrows
		Map<String, String> performPostRequestToKeycloak(MultiValueMap<String, String> formData) {
			RestClient restClient = RestClient.create();
			URI authorizationURI = new URIBuilder(keycloak.getAuthServerUrl() + "/realms/by-kiel-dev/protocol/openid-connect/token").build();
			var response = restClient.post().uri(authorizationURI)
					.contentType(MediaType.APPLICATION_FORM_URLENCODED)
					.body(formData);
			return response.retrieve().body(Map.class);

		}

	}
}