package de.ozgcloud.admin.settings.postfach;

import static org.assertj.core.api.Assertions.*;
import static org.springframework.security.test.web.servlet.request.SecurityMockMvcRequestPostProcessors.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import java.util.List;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.data.rest.RepositoryRestProperties;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.http.MediaType;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;

import com.fasterxml.jackson.databind.ObjectMapper;

import de.ozgcloud.admin.settings.Settings;
import de.ozgcloud.admin.settings.SettingsBody;
import de.ozgcloud.admin.settings.SettingsConstants;
import de.ozgcloud.admin.settings.SettingsTestFactory;
import de.ozgcloud.common.test.DataITCase;
import lombok.SneakyThrows;

@DataITCase
@AutoConfigureMockMvc
@WithMockUser
class PostfachITCase {
	private static final String POSTFACH_NAME = "Postfach";

	@Autowired
	private MockMvc mockMvc;

	@Autowired
	private MongoOperations mongoOperations;

	@Autowired
	private RepositoryRestProperties restProperties;

	private Settings settingsWithPostfach = SettingsTestFactory.createBuilder()
			.name(POSTFACH_NAME)
			.settingsBody(PostfachTestFactory.create())
			.build();

	@BeforeEach
	void clear() {
		mongoOperations.dropCollection(Settings.class);
	}

	@Nested
	class TestSave {

		@Test
		@SneakyThrows
		void shouldHaveResponseStatusCreated() {
			var result = performPost(settingsWithPostfach);

			result.andExpect(status().isCreated());
		}

		@Test
		@SneakyThrows
		void shouldCreateSettingsItem() {
			performPost(settingsWithPostfach);

			assertThat(getSettingWithPostfachFromDb())
					.usingRecursiveComparison().ignoringFields("id").isEqualTo(settingsWithPostfach);
		}

		@Nested
		class TestPostfachSettings {

			@Test
			@SneakyThrows
			void shouldBeInstanceOfPostfach() {
				performPost(settingsWithPostfach);

				assertThat(getSettingWithPostfachFromDb()
						.getSettingsBody()).isInstanceOf(Postfach.class);
			}

			@Test
			@SneakyThrows
			void shouldCreateEmptySignatur() {
				var postfachWithEmptySignatur = PostfachTestFactory.createBuilder()
						.signatur(StringUtils.EMPTY).build();
				var settingsWithPostfachEmptySignatur = createSettingsWithPostfach(postfachWithEmptySignatur);

				performPost(settingsWithPostfachEmptySignatur);

				assertThat(getPostfachFromDb().getSignatur()).isEmpty();
			}

			private Postfach getPostfachFromDb() {
				return (Postfach) getSettingWithPostfachFromDb().getSettingsBody();
			}

			@Nested
			class TestAbsenderValidation {
				@Test
				@SneakyThrows
				void shouldReturnUnprocessableEntityOnEmptyName() {
					var absenderWithEmptyName = AbsenderTestFactory.createBuilder().name(StringUtils.EMPTY).build();
					var settingsWithPostfachEmptyAbsenderName = createSettingsWithPostfachAbsender(absenderWithEmptyName);

					var result = performPost(settingsWithPostfachEmptyAbsenderName);

					result.andExpect(status().isUnprocessableEntity());
				}

				@Test
				@SneakyThrows
				void shouldReturnUnprocessableEntityOnEmptyAnschrift() {
					var absenderWithEmptyAnschrift = AbsenderTestFactory.createBuilder().anschrift(StringUtils.EMPTY).build();
					var settingsWithPostfachEmptyAnschrift = createSettingsWithPostfachAbsender(absenderWithEmptyAnschrift);

					var result = performPost(settingsWithPostfachEmptyAnschrift);

					result.andExpect(status().isUnprocessableEntity());
				}

				@Test
				@SneakyThrows
				void shouldReturnUnprocessableEntityOnEmptyDienst() {
					var absenderWithEmptyDienst = AbsenderTestFactory.createBuilder().dienst(StringUtils.EMPTY).build();
					var settingsWithPostfachEmptyDienst = createSettingsWithPostfachAbsender(absenderWithEmptyDienst);

					var result = performPost(settingsWithPostfachEmptyDienst);

					result.andExpect(status().isUnprocessableEntity());
				}

				@Test
				@SneakyThrows
				void shouldReturnUnprocessableEntityOnEmptyMandant() {
					var absenderWithEmptyMandant = AbsenderTestFactory.createBuilder().mandant(StringUtils.EMPTY).build();
					var setttingsWithPostfachEmptyMandant = createSettingsWithPostfachAbsender(absenderWithEmptyMandant);

					var result = performPost(setttingsWithPostfachEmptyMandant);

					result.andExpect(status().isUnprocessableEntity());
				}

				@Test
				@SneakyThrows
				void shouldReturnUnprocessableEntityOnEmptyGemeindeSchluessel() {
					var absenderWithEmptyGemeindeschluessel = AbsenderTestFactory.createBuilder().gemeindeschluessel(StringUtils.EMPTY).build();
					var settingsWithPostfachEmptyGemeindeschluessel = createSettingsWithPostfachAbsender(absenderWithEmptyGemeindeschluessel);

					var result = performPost(settingsWithPostfachEmptyGemeindeschluessel);

					result.andExpect(status().isUnprocessableEntity());
				}

				private Settings createSettingsWithPostfachAbsender(Absender absender) {
					var postfach = PostfachTestFactory.createBuilder().absender(absender).build();
					return createSettingsWithPostfach(postfach);
				}

			}

			private Settings createSettingsWithPostfach(SettingsBody postfach) {
				return SettingsTestFactory.createBuilder().name(POSTFACH_NAME).settingsBody(postfach).build();
			}
		}

		@SneakyThrows
		private ResultActions performPost(Settings setting) {
			var postBody = convertSettingToString(setting);
			return mockMvc.perform(post(String.join("/", restProperties.getBasePath(),
					SettingsConstants.PATH))
							.with(csrf())
							.contentType(MediaType.APPLICATION_JSON)
							.content(postBody));
		}
	}

	@Nested
	class TestGet {
		private String id;
		@Autowired
		private ObjectMapper mapper;

		@BeforeEach
		void init() {
			id = mongoOperations.save(settingsWithPostfach).getId();
		}

		@Test
		@SneakyThrows
		void shouldHaveStatusOkForExisting() {
			var result = performGet();

			result.andExpect(status().isOk());
		}

		@Test
		@SneakyThrows
		void shouldHaveInstanceOfPostfach() {

			var result = performGet();

			assertThat(mapper.readValue(result.andReturn().getResponse().getContentAsString(), Settings.class).getSettingsBody())
					.isInstanceOf(Postfach.class);
		}

		@SneakyThrows
		private ResultActions performGet() {
			return mockMvc.perform(get(String.join("/", restProperties.getBasePath(), SettingsConstants.PATH, id)));
		}
	}

	@Nested
	class TestPut {
		private String id;
		private Postfach updatedPostfach = PostfachTestFactory.createBuilder()
				.absender(AbsenderTestFactory.createBuilder()
						.name("Neuer Name")
						.anschrift("Neue Anschrift")
						.build())
				.build();
		private Settings updatedSettings = SettingsTestFactory.createBuilder()
				.name(POSTFACH_NAME)
				.settingsBody(updatedPostfach)
				.build();

		@BeforeEach
		void init() {
			id = mongoOperations.save(settingsWithPostfach).getId();
		}

		@Test
		@SneakyThrows
		void shouldHaveStatusNoContent() {
			var result = performPut();

			result.andExpect(status().isNoContent());
		}

		@Test
		@SneakyThrows
		void shouldHaveUpdatedSettings() {
			performPut();

			assertThat(getSettingWithPostfachFromDb())
					.usingRecursiveComparison().ignoringFields("id").isEqualTo(updatedSettings);
		}

		@Test
		@SneakyThrows
		void shouldHaveExactlyOnePostfachSetting() {
			performPut();

			List<Settings> postfachSettings = mongoOperations.find(createQueryForPostfach(), Settings.class);

			assertThat(postfachSettings).hasSize(1);
		}

		@SneakyThrows
		private ResultActions performPut() {
			var body = convertSettingToString(updatedSettings);
			return mockMvc.perform(put(String.join("/", restProperties.getBasePath(), SettingsConstants.PATH, id))
					.with(csrf())
					.contentType(MediaType.APPLICATION_JSON)
					.content(body));
		}

	}

	private Settings getSettingWithPostfachFromDb() {
		return mongoOperations.findOne(createQueryForPostfach(), Settings.class);
	}

	private Query createQueryForPostfach() {
		return new Query().addCriteria(Criteria.where("name").in(POSTFACH_NAME));
	}

	@SneakyThrows
	private String convertSettingToString(Settings setting) {
		return SettingsTestFactory.buildSettingsJson(setting, PostfachTestFactory.buildPostfachJson((Postfach) setting.getSettingsBody()));
	}
}
